val merge : Constrexpr.constr_expr_r CAst.t ->
            Constrexpr.constr_expr_r CAst.t ->
            Constrexpr.constr_expr_r CAst.t ->
            unit

val merge_test : Constrexpr.constr_expr_r CAst.t ->
                 unit
