(** * Bib: Bibliography *)

(* ################################################################# *)
(** * Resources cited in this volume *)

(**

[Bertot 2004] Interactive Theorem Proving and Program Development:
  Coq'Art: The Calculus of Inductive Constructions, by Yves Bertot and
  Pierre Casteran.  Springer-Verlag, 2004.
  http://tinyurl.com/z3o7nqu

[Chlipala 2013] Certified Programming with Dependent Types, by
  Adam Chlipala.  MIT Press.  2013.  http://tinyurl.com/zqdnyg2

[Lipovaca 2011] Learn You a Haskell for Great Good! A Beginner's
  Guide, by Miran Lipovaca, No Starch Press, April 2011.
  http://learnyouahaskell.com

[O'Sullivan 2008] Bryan O'Sullivan, John Goerzen, and Don Stewart:
  Real world Haskell - code you can believe in. O'Reilly
  2008. http://book.realworldhaskell.org

[Pugh 1991] Pugh, William. "The Omega test: a fast and practical
  integer programming algorithm for dependence analysis." Proceedings
  of the 1991 ACM/IEEE conference on Supercomputing. ACM, 1991.
  http://dl.acm.org/citation.cfm?id=125848

[Wadler 2015] Philip Wadler. "Propositions as types."
  Communications of the ACM 58, no. 12 (2015): 75-84.
  http://dl.acm.org/citation.cfm?id=2699407

*)

(* $Date: 2017-05-23 13:45:44 -0400 (Tue, 23 May 2017) $ *)
